// !Contents v2.23

// Produces a contents page by scanning through all the text and looking
// for headings of particular styles and formatting appropriately.
// Originally written by Jamie Lawrence, 1997

// Modified by Roger Noble (rgn), January - April 1999 et seq.

// Version History:
// v0.01 (8th April 1997)   :  Initial development version - does nothing
// v0.02 (20th May 1997)    :  Full editing of the styles that !Contents will
//                             use.  Sorted out all the windows and menu
//                             buttons. Setting of bookmarks. due to style
//                             name stripping can't have style names with
//                             spaces.
//       (25th May 1997)    :  Actually produces the contents pagee.  Chapter
//                             numbers don't work yet.  Much relience on the
//			       user to produce the contents styles and tabs.
//
// v1.0 (31st January 1999):   Modifications by rgn.
//			       Fixed up spaces in styles.
//			       Now handles multiple chapters.
//			       Tidied up bookmark handling. Now recognises when contents
//			       bookmark is already set in document.
//			       Added event handler to close style dialogue when file closed.
//                             Added user-selectable options for inserting "Page"
//			       and chapter numbers.
//			       Added user-selectable options for "Page" string and
//			       chapter number separator.
//			       Added option to set chapter number offset.
//                             Added saving of selected options and styles with document.
// v1.01 (23rd February 1999): Forgot to add on document starting chapter number.
// v1.02 (16th March 1999):    Changed global variable names to avoid any clashes.
//			       Used extra temporary bookmark for scanning through story
//			       in an attempt to nail the looping contents generation
//			       problem.
//			       Removed (temporarily) the trap for looping contents
//			       generation. Not much hope though ...
// v2.00 (7th April 1999):     Finally got tired of looping contents generation so re-wrote
//			       it to use a scrap file to save the contents headings (hence
//			       avoiding a lot of jumping backwards and forwards with bookmarks).
//			       This appears to have fixed the looping problem. It works faster
//			       too.
// v2.01 (8th April 1999):     Added support for inserted chapter numbers and active dates in
//			       headings.
// v2.02 (6th May 1999):       Replaced Contents_isstyle2 with a single call of
//			       iscurrentstyle(0) in Contents_isstyle. Should make it go a
//			       little faster.
// v2.03 (18th May 1999):      In some (but not all) cases when there was no newline at the end
//			       of a heading, attempting to move a bookmark to the next
//			       character did not work, resulting in an infinite loop when
//			       copying it. Added a check for end of text.
//			       Also fixed similar problem that happened if a frame overflowed.
// v2.04 (21st May 1999):      Trapped for < and { in contents headings and made sure that they
//			       were not being treated as system variable and macro expansions.
//			       Did a few tweaks to the contents writing to stop if getting
//			       confused when the starting position is followed by things like
//			       a forced new page.
// v2.05 (11th Sept 1999):     When a new view was opened the current file was not being set,
//			       resulting in reading either incorrect or totally dud tag data.
//			       The latter caused a type 5 error with RO4.
//			       Fixed by David Pilling before I got there ... (thanks DP!)
// v2.10 (13th October 1999):  Modified user-interface: replaced separate menu items with a
//			       single window which is also used to set the styles and other
//			       options.
//			       Removed separate Save action and replaced with automatic save
//			       whenever contents are produced.
//			       Added Set Values action.
//			       Fixed bug which gave incorrect ticked styles in style menus.
//			       Put in a trap to ensure that the view is always set to one for
//			       the current file (e.g., required when menu is clicked over a
//			       view other than the current one).
//			       Page and Chapter number options were being left set if Cancel was
//			       clicked.
//			       Tweaked-up the code that attempts to stop the contents writing
//			       getting confused when starting directly before existing text.
// v2.11 (30th November 1999)  Added macro definition for doing contents.
// v2.12 (9th December 1999)   Added trap for contents headings being too long. Too big a string
//			       appears to produce a type 5 error in type().
//			       Fixed failure to load new document information when a new file
//			       was loaded after a previous one had been closed.
// v2.13 (4th February 2000)   Added support for page numbers in Roman format. This limits the
//			       use to OvP 2.57 or later.
// v2.14 (19th March 2000)     Fixed faulty fading of page number writable icon.
//			       Do contents from cursor was always starting at beginning of the
//			       chapter; fixed now.
// v2.15 (11th August 2000)    Removed unused variable ChapterMax.
// v2.20 (23rd March 2001)     Added option for printing chapter titles in contents listing.
// v2.21 (6th April 2001)      Display hour glass when busy.
// v2.22 (13th June 2002)      Do contents from cursor now just does the contents from the current
//                             cursor location up to the end of the last chapter.
// v2.23 (12th July 2003)      Added support for specifying a different style for page numbers.
// v2.23 (5th July 2005)   Windows version by David Pilling


//=============================================================================================

// Global Variables.

// Options.

int Contents_ChaptNum = 0;    // Print chapter numbers before page numbers?
                              // 1 for yes, 0 for no
int Contents_PageOpt = 0;     // Print "Page" before page numbers, 1 => yes.

string Contents_Page;         // The string printed before the page numbers
string Contents_ChaptPageSep; // The string to seperate the chapter & page nums

int Contents_ChapterOffset;   // Chapter number offset (added to chapter numbers)

int Contents_ChapterTitles = 0;    // Print chapter titles in the contents listing:
                                   // 1 for yes, 0 for no
int Contents_ChapterTitlesNum = 0; // Print chapter numbers before the titles:
                                   // 1 for yes, 0 for no

int Contents_ChapterTitlesPag = 0; // Print page numbers after chapter titles:
                                   // 1 for yes, 0 for no

int Contents_ChaptNumTemp;    // Used to hold values whilst Contents dialogue active
int Contents_PageOptTemp;
int Contents_ChapterTitlesTemp;
int Contents_ChapterTitlesNumTemp;
int Contents_ChapterTitlesPagTemp;

string Contents_ChapterTitlesSep;
string Contents_ChapterTitlesSepAct;

// Contents scrap file name.


// Handles and techie things.

int Contents_stylemenu;     // Style menu-button menu handle
int Contents_win;           // The Edit Style Window handle

// Styles.

int Contents_Style1 = -1;   // Style handles
int Contents_Style2 = -1;
int Contents_Style3 = -1;
int Contents_Style4 = -1;
int Contents_StyleT = -1;
string Contents_Style1s;    // Style names
string Contents_Style2s;
string Contents_Style3s;
string Contents_Style4s;
string Contents_StyleTs;
int Contents_temp1 = -1;    // Used in Contents dialogue to hold style handles
int Contents_temp2 = -1;
int Contents_temp3 = -1;
int Contents_temp4 = -1;
int Contents_tempT = -1;

// Window icons.

int Contents_current_icon = 0; // Which style am I currently editing
                               // Bad style to have it global, but this is the simplest
                               // way of getting the job done
// Bookmarks.

int Contents_BM;       // Where the contents will be produced
int Contents_BMend;    // The end of the current contents page

int Contents_first;    // Used to force Contents_BM back to ths start of the contents

// Document-related stuff.

int Contents_current_file = 0; 			       // Handle to current file
int Contents_current_view = 0;			       // Handle to current view

// Constants.

int Contents_tag = ('c'<<24) | ('o'<<16) | ('n'<<8);   // Base tag value
int Contents_truncate = 512;			       // Truncate a heading if longer than this

//=============================================================================================

////////////////////////////////////////////////
// Contents Page Producing Functions          //
////////////////////////////////////////////////

// Test if the style is one of the styles for the contents page. Returns the
// style number, or zero if not. Returns -1 for the titles style.

int Contents_isstyle()
{
  caretcontext();

  iscurrentstyle(0);			      // Set up styles

  if (iscurrentstyle(Contents_Style1) != 0)   // Is it style 1 at the caret?
    return(1);                                // If so return 1
  if (iscurrentstyle(Contents_Style2) != 0)   // and so on
    return(2);
  if (iscurrentstyle(Contents_Style3) != 0)
    return(3);
  if (iscurrentstyle(Contents_Style4) != 0)
    return(4);
  if (iscurrentstyle(Contents_StyleT) != 0)
    return(-1);

  return(0);                       // It wasn't any of the known styles
}

//---------------------------------------------------------------------------------------------

// Clear all the styles from the selected or current text.

int Contents_clearallstyles()
{
 int i = 0;
 string s;

 while(getcurrentstyle(i, s, 2) != 0)
 {
   remstyle(i++);
 }

 return(0);
}

//---------------------------------------------------------------------------------------------

// Find a style with the name "stylename", return the handle for it. Returns -1
// if the style is not found.

int Contents_findstyle(string stylename)
{
  int temp;
  string tempstring;

  for (temp = 0; getuserstyle(temp, tempstring, 0) != 0; temp++)
  {
    if (tempstring == stylename)
      return(temp);
  }

  return -1;
}

//---------------------------------------------------------------------------------------------

// Sets the style handles from the style names.

void Contents_set_styles()
{
    Contents_Style1 = Contents_findstyle (Contents_Style1s);
    Contents_Style2 = Contents_findstyle (Contents_Style2s);
    Contents_Style3 = Contents_findstyle (Contents_Style3s);
    Contents_Style4 = Contents_findstyle (Contents_Style4s);
    Contents_StyleT = Contents_findstyle (Contents_StyleTs);
    Contents_temp1 = Contents_Style1;
    Contents_temp2 = Contents_Style2;
    Contents_temp3 = Contents_Style3;
    Contents_temp4 = Contents_Style4;
    Contents_tempT = Contents_StyleT;
}

//---------------------------------------------------------------------------------------------

// Builds the contents page for one story. Returns zero if there are more
// chapters (stories) to do.

int Contents_styletest (int tempbm, int cont_file)
{
  int level;
  int i = 1;
  int end;
  int bm;
  int bm1;
  int bm2;
  int c;
  int chapter;
  int numberformat;
  int done_title = !Contents_ChapterTitles;

  bm = bmcreate ("Contents_SBM");        // Bookmark for going through current story
  setbmtobm(bm, tempbm);
  bm1 = bmcreate ("Contents_SBM1");	 // Bookmark for going through story
  bm2 = bmcreate ("Contents_SBM2");	 // Bookmark for checking

  chapter = bmchapter(tempbm) + chapterstart(currentfile());

  // Obtain the chapter number format. This is not supported for versions of OvP prior
  // to 2.57. We could test for the version number here, but since the earlier versions
  // don't have itors() the applet will not load properly with these anyway.

  numberformat = (chapterbits(bmchapter(tempbm), currentfile()) >> 4) & 0xF;

  while (i != 0)                         // While not the end of text
  {
    setcarettobm(bm);                    // Move the caret to the bookmark
    setbmtobm(bm1,bm);			 // Bookmark at start of selected text
    i = bmmove(bm, 1, 3);             	 // Move bm to end of paragraph
    setzonetobm(bm);                 	 // Select the bit between caret & bm
    level = -1;
    level = Contents_isstyle();          // Find out what style it is, and
                                         // do I need it

    if (cancopy() != 0 && level != 0)    // If I can copy this selection and
                                         // its a style I need
    {
      // Got something to copy to the contents file. First check if we are to
      // write it. We write it if it is a normal section heading, or if it is
      // a chapter title and we are doing them and we have not already written it.

      if (level > 0 || (level < 0 && !done_title))
      {
        if (level < 0) done_title = 1;

        // First write the contents level, chapter and page numbers, and the
        // chapter number format. If

        filewritei(level, cont_file);
        filewritei(chapter, cont_file);
        filewritei(bmpage(bm) + pagestart(bmchapter(tempbm),currentfile()), cont_file);
        filewritei(numberformat, cont_file);

        // Now the contents text.

        do
        {
          c = bmchar(bm1);

          // Test for special characters.

    	  switch (c)
    	  {
    	    case 18:			 // Tab
    	      fileputc ('\t', cont_file);
    	      break;
    	    case 4:			// Current chapter number
    	      filewrites (itos(chapter), cont_file);
    	      break;
    	    case 5:			// Current date
    	      filewrites ("{activedate}", cont_file);
    	      break;
    	    case '<':			// <, replace with |< to avoid system variable expansion
    	      filewrites ("|<", cont_file);
    	      break;
    	    case '{':			// {, replace with |{ to avoid macro expansion
    	      filewrites ("|{", cont_file);
    	      break;
    	    default:			// Other characters
    	      if (c >= 32)
    	        fileputc (c, cont_file);
    	      break;
    	  }
    	  setbmtobm (bm2, bm1);;		// Remember where we are
          end = bmmove(bm1, 1, 0);         	// Next character
    	  if (bmcmp(bm1,bm2) <= 0) end = 0;	// If we did not move, finish
        }
        while (end != 0 && bmcmp(bm1,bm) < 0)
        fileputc ('\n', cont_file);	 // To terminate text
      }
    }

    // Next paragraph.

    setbmtobm (bm2, bm);			// Remember where we are
    i = bmmove(bm, 1, 7);       		// Move down a paragraph
    if (i != 0 && bmcmp(bm,bm2) <= 0) i = 0;	// If we did not move, finish story
  }

  bmdelete (bm2);
  bmdelete (bm1);
  bmdelete (bm);

  i = bmnextstory(tempbm);		 // Move to next story

  return i;
}

//---------------------------------------------------------------------------------------------

// Do the actual contents.

void Contents_write_contents (int cont_file)
{
  int level;
  int chapter;
  int page;
  string heading;
  int len;
  int truncate_count= 0;
  int numberformat;
  int styleh;
  int pagebm;
  int pagebm1;

  fileseek (0, cont_file);		// Reposition contents file
  setcarettobm (Contents_BM);		// Move to contents start
  type("|M");				// Start a new line just in case.
  bmmove (Contents_BM, 0, 7);		// Back up to before it
  setcarettobm(Contents_BM);
  setbmtocaret(Contents_BMend);
  bmmove (Contents_BMend, 1, 2);	// Get rid of any existing styles and effects
  setzonetobm (Contents_BMend);
  cleareffects(0);
  Contents_clearallstyles();
  setbmtobm(Contents_BMend, Contents_BM);
  pagebm = bmcreate("Contents_BMpage");
  pagebm1 = bmcreate("Contents_BMpage1");

  // Go through the file and write the contents.

  while(!fileeof(cont_file))
  {
    level = filereadi(cont_file);	// Contents level

    if (level >= 1 && level <= 4 || level == -1)
    {
      chapter = filereadi(cont_file);	   // Chapter number
      page = filereadi(cont_file);	   // Page number
      numberformat = filereadi(cont_file); // Page number format
      filereads (heading, cont_file);	   // Section heading

      len = slen(heading);


      // Add the appropriate contents style.

      cleareffects(0);
      Contents_clearallstyles();

      switch(level)                     // Add the appropriate contents style
      {
        case -1:
          styleh = Contents_findstyle("Contents Heading");
          if (styleh >= 0) addstyle(styleh);
          break;

        case 1:
          styleh = Contents_findstyle("Contents1");
          if (styleh >= 0) addstyle(styleh);
          break;

        case 2:
          styleh = Contents_findstyle("Contents2");
          if (styleh >= 0) addstyle(styleh);
          break;

        case 3:
          styleh = Contents_findstyle("Contents3");
          if (styleh >= 0) addstyle(styleh);
          break;

        case 4:
          styleh = Contents_findstyle("Contents4");
          if (styleh >= 0) addstyle(styleh);
          break;

        default:
          break;
      }

      // Write the heading to the contents, missing out the \n at its end.

      if (slen(heading) > 1)
      {
        if (len > Contents_truncate+1)     // Check for heading too long (too big and we
        {				   // get a Type 5 error on the type line below ...
          truncate_count++;
          len = Contents_truncate+1;
        }

        // If we are doing a chapter title then insert the chapter number if
        // requested.

        if (level == -1 && Contents_ChapterTitlesNum)
        {
          type(itos(chapter+Contents_ChapterOffset));
          type(Contents_ChapterTitlesSepAct);
        }

        type(mids(heading,0,len-1));       // Write the heading

        if (Contents_first)		   // Get the contents start bookmark back to start
        {
          bmmove (Contents_BM, 0, 3);
          Contents_first = 0;
        }

        setcarettobm(Contents_BMend);      // Move caret to end of contents

        // Add the chapter/page number for normal entries, or if requested when doing a chapter title.

        if (level > 0 || Contents_ChapterTitlesPag)
        {
	  setbmtocaret(pagebm);
          type("{Tab}");                   // Add the tab in
          bmmove(pagebm, 0, 0);
          
          if (Contents_PageOpt)		   // If "Page" wanted before page numbers
            type(Contents_Page);
          if (Contents_ChaptNum)           // If chapter numbers wanted
          {
            type(itos(chapter+Contents_ChapterOffset)); // Print the chapter number allowing for the offset
            type(Contents_ChaptPageSep);                // Print the chapter/page seperator
          }

          if (numberformat == 0)		 // Print the page number ...
            type(itors(page));	  	 // as Roman upper case
          else if (numberformat == 1)
            type(-itors(page));		 // as Roman lower case
          else
            type(itos(page));		 // as decimal
        }

        type("|M");                      // Type in a return
        
        // Set the page number style if any.
        
        styleh = Contents_findstyle("Contents PageNum");
        if (styleh >= 0)
        {
          // This horribly convoluted way of doing it is necessary because
          // if the style is just added at the right space and then remstyled
          // at its end, in practice it is not removed. So we back up to the
          // previous line, apply the page number style to the tab and page number
          // and then move back to the start of our new line. Any other method
          // seems to leave the style stuck on.
          
          setbmtocaret(pagebm1);
          bmmove(pagebm1, 0, 0);
          setcarettobm(pagebm1);
          setzonetobm(pagebm);
          addstyle(styleh);
          bmmove(Contents_BMend, 1, 2);    // Move end bm to next contents line
          clearselection();
          setcarettobm(Contents_BMend);
        }
        else
          bmmove(Contents_BMend, 1, 2);    // Move end bm to next contents line
      }
    }
  }
  
  bmdelete(pagebm1);
  bmdelete(pagebm);

  type ("{Deletef}");			// Remove the new line we inserted
  
  // Did we truncate any headings?

  if (truncate_count > 0)
  {
    string s = "{CONTENTS_11}";
    translate(s);
    messagebox(itos(truncate_count) + s);
  }
}

//---------------------------------------------------------------------------------------------

// Produce the contents.

void Contents_do_contents (int tempbm)
{
  int    cont_file;
  int    cont_scrap;
  string Contents_file;


  Contents_set_styles();
  Contents_first = 1;			        // First time round
  Contents_BMend = bmcreate("Contents_End");	// Marks current end of contents
  setbmtobm(Contents_BMend,Contents_BM);	// Where contents start

  cont_scrap=scrapcreate();
  scrapname(cont_scrap,Contents_file);

  cont_file=fileopen(Contents_file,"wb+");
  if (cont_file == 0)
  {
    string s = "{CONTENTS_07}";
    translate(s);
    messagebox(s);
  }
  else
  {
//    swi13(0x406c0, 0, 0, 0);
    while (!Contents_styletest(tempbm, cont_file));
    Contents_write_contents(cont_file);
    fileclose(cont_file);
    systems("Del " + Contents_file);		// Delete scratch file
    bmdelete (Contents_BMend);
    bmdelete (tempbm);
    setcarettobm (Contents_BM);	        	// Move back to start of contents
//    swi13(0x406c1, 0, 0, 0);
  }

  scrapremove(cont_scrap);

}

//=============================================================================================

////////////////////////////////////////////////
//  Contents Style Menu Functions             //
////////////////////////////////////////////////

// Checks if any contents styles have been set and that there is a contents bookmark.
// Returns 1 if any styles have been set, 0 otherwise.

int Contents_check_ready()
{
  // Check if the contents titles options can be set.

  if (Contents_tempT <= -1)
  {
    set_icon_state (Contents_win, 33, 0x00400000, 0x00400000);
    set_icon_state (Contents_win, 37, 0x00400000, 0x00400000);
    set_icon_state (Contents_win, 38, 0x00400000, 0x00400000);
    set_icon_state (Contents_win, 40, 0x00400000, 0x00400000);
  }
  else
  {
    set_icon_state(Contents_win, 33, 0x00000000, 0x00400000);
    if (Contents_ChapterTitlesTemp)
    {
      set_icon_state (Contents_win, 37, 0x00000000, 0x00400000);
      set_icon_state (Contents_win, 40, 0x00000000, 0x00400000);
      if (Contents_ChapterTitlesNumTemp)
        set_icon_state (Contents_win, 38, 0x00000000, 0x00400000);
      else
        set_icon_state (Contents_win, 38, 0x00400000, 0x00400000);
    }
    else
    {
      set_icon_state (Contents_win, 37, 0x00400000, 0x00400000);
      set_icon_state (Contents_win, 38, 0x00400000, 0x00400000);
    }
  }

  if (Contents_temp1 <= -1 && Contents_temp2 <= -1 && Contents_temp3 <= -1
      && Contents_temp4 <= -1)
  {
    // No contents styles are set. Disable the Do Contents actions.

    set_icon_state (Contents_win, 29, 0x00400000, 0x00400000);
    set_icon_state (Contents_win, 0, 0x00400000, 0x00400000);

    return 0;
  }
  else
  {
    int tempbm;

    // Check if there is a location for the contents.

    if (tempbm=bmfind("Contents_Start"))	// Is there a contents bookmark?
      set_icon_state (Contents_win, 29, 0x00000000, 0x00400000);
    else
      // No, disable the Do Contents from Cursor action.
      set_icon_state (Contents_win, 29, 0x00400000, 0x00400000);

    set_icon_state (Contents_win, 0, 0x00000000, 0x00400000);

    return 1;
  }
}

//---------------------------------------------------------------------------------------------

int Contents_choosestyle(void)
{
  int i = 0;
  string tempstring = "";

  if(Contents_stylemenu)
  {
    delete_menu(Contents_stylemenu);
    Contents_stylemenu = create_menu("{CONTENTS_05}");
  }

  // Add all the menu entries.

  addentry_menu (Contents_stylemenu, "Contents_styleentry", "Contents_styleflags", "", "", "Clear Style");

  for(i=0; getuserstyle(i, tempstring, 0) != 0; i++)
  {
    addentry_menu(Contents_stylemenu, "Contents_styleentry", "Contents_styleflags", "", "", tempstring);
  }
}

//---------------------------------------------------------------------------------------------

int Contents_styleentry (int n, int subcode)
{
  string stylename;

  if (n > 0)
    getuserstyle (n-1, stylename, 0);
  else
    stylename = "";		// Clear style

  switch (Contents_current_icon)
  {
    case 6:
      writeicon (Contents_win, 102, stylename);
      Contents_temp1 = n-1;
      break;

    case 7:
      writeicon (Contents_win, 3, stylename);
      Contents_temp2 = n-1;
      break;

    case 8:
      writeicon (Contents_win, 4, stylename);
      Contents_temp3 = n-1;
      break;

    case 9:
      writeicon (Contents_win, 5, stylename);
      Contents_temp4 = n-1;
      break;

    case 36:
      writeicon (Contents_win, 35, stylename);
      Contents_tempT = n-1;
      Contents_check_ready();
      break;

    default:
      return (0);
      break;
  }

  Contents_check_ready();	// Check if ready to do contents now

  return (1);
}

//---------------------------------------------------------------------------------------------

// Tick, shade etc the items appropriately.

int Contents_styleflags (int n, string &s)
{
  int bits = 0;

  switch (Contents_current_icon)   // Find out which icon we are editing
  {
    case 6:
      if(Contents_temp1 == n-1)    // is the current style = this menu item
        bits = TICKED;     // if so then tick it
      break;

    case 7:
      if(Contents_temp2 == n-1)    // and so on...
        bits = TICKED;
      break;

    case 8:
      if(Contents_temp3 == n-1)
        bits = TICKED;
      break;

    case 9:
      if(Contents_temp4 == n-1)
        bits = TICKED;
      break;

    case 36:
      if(Contents_tempT == n-1)
        bits = TICKED;
      break;

    default:
      return (0);
      break;
  }

  return (bits);   // Return the flags code
}

//=============================================================================================

////////////////////////////////////////////////
// Style Editing Window Code                  //
////////////////////////////////////////////////

// Load the default options.

void Contents_load_defaults()
{
  Contents_Style1 = -1;
  Contents_Style2 = -1;
  Contents_Style3 = -1;
  Contents_Style4 = -1;
  Contents_StyleT = -1;
  Contents_temp1 = -1;
  Contents_temp2 = -1;
  Contents_temp3 = -1;
  Contents_temp4 = -1;
  Contents_tempT = -1;
  Contents_Style1s = "";
  Contents_Style2s = "";
  Contents_Style3s = "";
  Contents_Style4s = "";
  Contents_StyleTs = "";
  Contents_ChaptNum = 0;
  Contents_PageOpt = 0;
  Contents_Page = "Page ";          // The string printed before the page numbers
  Contents_ChaptPageSep = ".";      // The string to separate the chapter & page nums
  Contents_ChapterOffset = 0;
  Contents_ChapterTitles = 0;
  Contents_ChapterTitlesNum = 0;
  Contents_ChapterTitlesSep = ". "; // The string used to separate chapter numbers from titles
  Contents_ChapterTitlesPag = 0;
}

//---------------------------------------------------------------------------------------------

// Save the current options with the document.

void Contents_save_options()
{
  tagwrites (Contents_current_file, Contents_tag+0, Contents_Style1s);
  tagwrites (Contents_current_file, Contents_tag+1, Contents_Style2s);
  tagwrites (Contents_current_file, Contents_tag+2, Contents_Style3s);
  tagwrites (Contents_current_file, Contents_tag+3, Contents_Style4s);
  tagwrite (Contents_current_file, Contents_tag+4, Contents_ChaptNum, 0);
  tagwrite (Contents_current_file, Contents_tag+5, Contents_PageOpt, 0);
  tagwrites (Contents_current_file, Contents_tag+6, Contents_Page);
  tagwrites (Contents_current_file, Contents_tag+7, Contents_ChaptPageSep);
  tagwrite (Contents_current_file, Contents_tag+8, Contents_ChapterOffset, 0);
  tagwrite (Contents_current_file, Contents_tag+9, Contents_ChapterTitles, 0);
  tagwrite (Contents_current_file, Contents_tag+10, Contents_ChapterTitlesNum, 0);
  tagwrites (Contents_current_file, Contents_tag+11, Contents_ChapterTitlesSep);
  tagwrite (Contents_current_file, Contents_tag+12, Contents_ChapterTitlesPag, 0);
  tagwrites (Contents_current_file, Contents_tag+13, Contents_StyleTs);
}

//---------------------------------------------------------------------------------------------

// Loads the saved options from the document. Returns true always.

int Contents_load_document()
{

  Contents_load_defaults();			// Set default values initially

  if (tagexists(Contents_current_file, Contents_tag+0, 0))
    tagreads (Contents_current_file, Contents_tag+0, Contents_Style1s);
  if (tagexists(Contents_current_file, Contents_tag+1, 0))
    tagreads (Contents_current_file, Contents_tag+1, Contents_Style2s);
  if (tagexists(Contents_current_file, Contents_tag+2, 0))
    tagreads (Contents_current_file, Contents_tag+2, Contents_Style3s);
  if (tagexists(Contents_current_file, Contents_tag+3, 0))
    tagreads (Contents_current_file, Contents_tag+3, Contents_Style4s);
  if (tagexists(Contents_current_file, Contents_tag+4, 0))
    Contents_ChaptNum = tagread (Contents_current_file, Contents_tag+4, 0);
  if (tagexists(Contents_current_file, Contents_tag+5, 0))
    Contents_PageOpt = tagread (Contents_current_file, Contents_tag+5, 0);
  if (tagexists(Contents_current_file, Contents_tag+6, 0))
    tagreads (Contents_current_file, Contents_tag+6, Contents_Page);
  if (tagexists(Contents_current_file, Contents_tag+7, 0))
    tagreads (Contents_current_file, Contents_tag+7, Contents_ChaptPageSep);
  if (tagexists(Contents_current_file, Contents_tag+8, 0))
    Contents_ChapterOffset = tagread (Contents_current_file, Contents_tag+8, 0);
  if (tagexists(Contents_current_file, Contents_tag+9, 0))
    Contents_ChapterTitles = tagread (Contents_current_file, Contents_tag+9, 0);
  if (tagexists(Contents_current_file, Contents_tag+10, 0))
    Contents_ChapterTitlesNum = tagread (Contents_current_file, Contents_tag+10, 0);
  if (tagexists(Contents_current_file, Contents_tag+11, 0))
    tagreads (Contents_current_file, Contents_tag+11, Contents_ChapterTitlesSep);
  if (tagexists(Contents_current_file, Contents_tag+12, 0))
    Contents_ChapterTitlesPag = tagread (Contents_current_file, Contents_tag+12, 0);
  if (tagexists(Contents_current_file, Contents_tag+13, 0))
    tagreads (Contents_current_file, Contents_tag+13, Contents_StyleTs);
  return 1;
}

//---------------------------------------------------------------------------------------------

// Parses the separator string provided and returns a string suitable for directly
// inserting into the text.

void Contents_parse_separator(string& separator, string& insert)
{
  int separator_len = slen(separator);
  int i = 0;
  int c;

  insert = "";

  while (i < separator_len)
  {
    c = schar(separator, i);

    if (c == '\\')
    {
      // Quoted character, check the next one.

      if (i+1 < separator_len)
      {
        c = schar(separator, i+1);

        switch (c)
        {
          case 't':    // Tab
            insert += "{tab}";
            break;

          case 'r':   // Newlines and the like, replace with space
          case 'n':
            insert += " ";
            break;

          default:    // Others, replace with the character following the '\'
            insert += chars(c);
            break;
        }

        i += 2;
      }
    }
    else if (c >= 32)
    {
      insert += chars(c);
      ++i;
    }
    else
      ++i;                     // Ignore control characters
  }
}

//---------------------------------------------------------------------------------------------

// Sets the current options from the dialogue box.

void Contents_set_options (int winhandle)
{
  Contents_Style1 = Contents_temp1;          // Permanently set the values for the styles
  Contents_Style2 = Contents_temp2;
  Contents_Style3 = Contents_temp3;
  Contents_Style4 = Contents_temp4;
  Contents_StyleT = Contents_tempT;
  if (Contents_Style1 >= 0)
    getuserstyle (Contents_Style1, Contents_Style1s, 0);
  else
    Contents_Style1s = "";
  if (Contents_Style2 >= 0)
    getuserstyle (Contents_Style2, Contents_Style2s, 0);
  else
    Contents_Style2s = "";
  if (Contents_Style3 >= 0)
    getuserstyle (Contents_Style3, Contents_Style3s, 0);
  else
    Contents_Style3s = "";
  if (Contents_Style4 >= 0)
    getuserstyle (Contents_Style4, Contents_Style4s, 0);
  else
    Contents_Style4s = "";
  if (Contents_StyleT >= 0)
    getuserstyle (Contents_StyleT, Contents_StyleTs, 0);
  else
    Contents_StyleTs = "";
  Contents_ChaptNum = Contents_ChaptNumTemp;
  Contents_PageOpt = Contents_PageOptTemp;
  Contents_ChapterTitles = Contents_ChapterTitlesTemp;
  Contents_ChapterTitlesNum = Contents_ChapterTitlesNumTemp;
  Contents_ChapterTitlesPag = Contents_ChapterTitlesPagTemp;
  Contents_Page = readicon (winhandle, 22);
  Contents_ChaptPageSep = readicon (winhandle, 21);
  Contents_ChapterOffset = stoi(readicon(winhandle, 25));
  Contents_ChapterTitlesSep = readicon (winhandle, 38);
}

//---------------------------------------------------------------------------------------------

// Display the contents style window.

int Contents_StyleWin()
{
  writeicon (Contents_win, 102, Contents_Style1s);
  writeicon (Contents_win, 3, Contents_Style2s);
  writeicon (Contents_win, 4, Contents_Style3s);
  writeicon (Contents_win, 5, Contents_Style4s);
  writeicon (Contents_win, 35, Contents_StyleTs);
  Contents_Style1 = Contents_findstyle(Contents_Style1s);
  Contents_Style2 = Contents_findstyle(Contents_Style2s);
  Contents_Style3 = Contents_findstyle(Contents_Style3s);
  Contents_Style4 = Contents_findstyle(Contents_Style4s);
  Contents_StyleT = Contents_findstyle(Contents_StyleTs);
  Contents_temp1 = Contents_Style1;
  Contents_temp2 = Contents_Style2;
  Contents_temp3 = Contents_Style3;
  Contents_temp4 = Contents_Style4;
  Contents_tempT = Contents_StyleT;
  Contents_PageOptTemp = Contents_PageOpt;
  Contents_ChaptNumTemp = Contents_ChaptNum;
  Contents_ChapterTitlesTemp = Contents_ChapterTitles;
  Contents_ChapterTitlesNumTemp = Contents_ChapterTitlesNum;
  Contents_ChapterTitlesPagTemp = Contents_ChapterTitlesPag;
  select_icon (Contents_win, 18, Contents_PageOpt);
  select_icon (Contents_win, 20, Contents_ChaptNum);
  select_icon (Contents_win, 33, Contents_ChapterTitles);
  select_icon (Contents_win, 37, Contents_ChapterTitlesNum);
  select_icon (Contents_win, 40, Contents_ChapterTitlesPag);
  writeicon (Contents_win, 22, Contents_Page);
  writeicon (Contents_win, 21, Contents_ChaptPageSep);
  writeicon (Contents_win, 25, itos(Contents_ChapterOffset));
  writeicon (Contents_win, 38, Contents_ChapterTitlesSep);
  if (Contents_PageOpt)
    set_icon_state (Contents_win, 22, 0x00000000, 0x00400000);
  else
    set_icon_state (Contents_win, 22, 0x00400000, 0x00400000);
  if (Contents_ChaptNum)
  {
    set_icon_state (Contents_win, 21, 0x00000000, 0x00400000);
    set_icon_state (Contents_win, 25, 0x00000000, 0x00400000);
    set_icon_state (Contents_win, 26, 0x00000000, 0x00400000);
    set_icon_state (Contents_win, 27, 0x00000000, 0x00400000);
  }
  else
  {
    set_icon_state (Contents_win, 21, 0x00400000, 0x00400000);
    set_icon_state (Contents_win, 25, 0x00400000, 0x00400000);
    set_icon_state (Contents_win, 26, 0x00400000, 0x00400000);
    set_icon_state (Contents_win, 27, 0x00400000, 0x00400000);
  }

  Contents_check_ready();

  display_window(Contents_win, 20, 1);

  return(1);
}

//---------------------------------------------------------------------------------------------

// Opens the menus off the buttons on the contents style window.

void Contents_openwinmenu (int winhandle, int icon)
{
  Contents_current_icon = icon;           // set current icon
  Contents_choosestyle();                 // make the menu
  open_menu(Contents_stylemenu);          // show the menu
}

//---------------------------------------------------------------------------------------------



void Contents_close(int handle)
{
 if(Contents_win)
 {
  close_window(Contents_win);
  delete_window(Contents_win);
  Contents_win=0;
  Contents_current_file=0;
  Contents_current_view=0;
 }
}







//  Mouse click event handler for the contents style window

void Contents_click (int winhandle, int icon, int bbits, int mx, int my)
{
  int tempbm;
  int tempbm1;
  int chapter;

  if (bbits != 4 && bbits != 1) return;	// Return if not left or right mouse button

  switch(icon)
  {
    case 0:                    // Default action window, produce the contents

      if (Contents_check_ready())
      {
        Contents_set_options(winhandle);
        Contents_save_options();

        // Check if the caret is anywhere.

        if (caretchar() >= 0)
        {
          // Obtain the actual separator to use for the contents number/titles separator.

          Contents_parse_separator(Contents_ChapterTitlesSep, Contents_ChapterTitlesSepAct);

          if (tempbm=bmfind("Contents_Start"))	// Is there already a contents bookmark?
            Contents_BM = tempbm;		// Yes, just use it
          else
            Contents_BM = bmcreate ("Contents_Start"); // No, create one.

          setbmtocaret (Contents_BM);		// Position contents start at caret

          if (tempbm=bmfind("Contents_End"))	// Delete any end of contents bookmark
            bmdelete (tempbm);

          tempbm = bmcreate("Contents_SFind");       // Used when searching through the text
          bmstartscan();
          bmnextstory(tempbm);
          bmmove(tempbm, 0, 4);			 // Make sure we are at the start of the text
          setcarettobm(tempbm);			 // Start at beginning of document
          Contents_do_contents(tempbm);

          if (bbits == 4) Contents_close(winhandle);
        }
        else
        {
          string s;
          s = "{CONTENTS_08}";
          translate(s);
          messagebox(s);
        }
      }
      break;

    case 2:                   // Cancel button  - closes window
      Contents_close(winhandle);
      break;

    case 29:		       // Do Contents from cursor.
      Contents_set_options(winhandle);
      Contents_save_options();

      // Check if the caret is anywhere.

      if (caretchar() >= 0)
      {
        if (tempbm=bmfind("Contents_Start"))	// Is there already a contents bookmark?
        {
          Contents_BM = tempbm;			// Yes, use it and its position

          if (tempbm=bmfind("Contents_End"))	// Delete any end of contents bookmark
            bmdelete (tempbm);

          tempbm = bmcreate("Contents_SFind");
          tempbm1 = bmcreate("Contents_SCurrent");
          bmstartscan();
          setbmtocaret(tempbm1);	        // Start at current cursor location
          chapter = bmchapter(tempbm1);         // Current chapter
          while (!bmnextstory(tempbm) && bmchapter(tempbm) != chapter);     // Get the scan into the correct chapter
          setbmtobm(tempbm, tempbm1);           // Starting location in chapter
          bmdelete(tempbm1);
          Contents_do_contents(tempbm);

          if (bbits == 4) Contents_close(winhandle);
        }
        else
        {
          string s = "{CONTENTS_06}";

          // There is no current contents bookmark, complain that we do not know where
          // to put the contents.

	  translate(s);
          messagebox(s);
        }
      }
      else
      {
        // There is no known cursor position.

        string s = "{CONTENTS_09}";
        translate(s);
        messagebox(s);
      }
      break;

    case 31:			// Set Contents Location
      if (tempbm=bmfind("Contents_Start"))	// Is there already a contents bookmark?
        Contents_BM = tempbm;			// Yes, just use it
      else
        Contents_BM = bmcreate ("Contents_Start"); // No, create one.

      setbmtocaret (Contents_BM);		// Position contents start at caret

      Contents_check_ready();

      if (tempbm=bmfind("Contents_End"))	// Delete any end of contents bookmark
        bmdelete (tempbm);

      break;

    case 32:			// Set Values
      Contents_set_options(winhandle);
      Contents_save_options();

      // Obtain the actual separator to use for the contents number/titles separator.

      Contents_parse_separator(Contents_ChapterTitlesSep, Contents_ChapterTitlesSepAct);

      if (bbits == 4) Contents_close(winhandle);
      break;

    case 6:                    // All the menu buttons
    case 7:
    case 8:
    case 9:
    case 36:
      // Open the menu off the menu buttons.
      Contents_openwinmenu (winhandle, icon);
      break;

    case 18:			// Page text option
      select_icon (winhandle, 18, Contents_PageOptTemp = !Contents_PageOptTemp);
      if (Contents_PageOptTemp)
        set_icon_state (Contents_win, 22, 0x00000000, 0x00400000);
      else
        set_icon_state (Contents_win, 22, 0x00400000, 0x00400000);
      break;

    case 20:			// Chapter number option
      select_icon (winhandle, 20, Contents_ChaptNumTemp = !Contents_ChaptNumTemp);
      if (Contents_ChaptNumTemp)
      {
        set_icon_state (Contents_win, 21, 0x00000000, 0x00400000);
        set_icon_state (Contents_win, 25, 0x00000000, 0x00400000);
        set_icon_state (Contents_win, 26, 0x00000000, 0x00400000);
        set_icon_state (Contents_win, 27, 0x00000000, 0x00400000);
      }
      else
      {
        set_icon_state (Contents_win, 21, 0x00400000, 0x00400000);
        set_icon_state (Contents_win, 25, 0x00400000, 0x00400000);
        set_icon_state (Contents_win, 26, 0x00400000, 0x00400000);
        set_icon_state (Contents_win, 27, 0x00400000, 0x00400000);
      }
      break;

    case 26:			// Bump down chapter number offset
    {
      int offset = stoi(readicon(Contents_win, 25));
      writeicon(Contents_win, 25, itos(--offset));
      break;
    }

    case 27:			// Bump up chapter number offset
    {
      int offset = stoi(readicon(Contents_win, 25));
      writeicon(Contents_win, 25, itos(++offset));
      break;
    }

    case 33:			// Chapter titles option
      select_icon (winhandle, 33, Contents_ChapterTitlesTemp = !Contents_ChapterTitlesTemp);
      if (Contents_ChapterTitlesTemp)
      {
        set_icon_state (Contents_win, 37, 0x00000000, 0x00400000);
        set_icon_state (Contents_win, 40, 0x00000000, 0x00400000);
        if (Contents_ChapterTitlesNumTemp)
          set_icon_state (Contents_win, 38, 0x00000000, 0x00400000);
        else
          set_icon_state (Contents_win, 38, 0x00400000, 0x00400000);
      }
      else
      {
        set_icon_state (Contents_win, 37, 0x00400000, 0x00400000);
        set_icon_state (Contents_win, 38, 0x00400000, 0x00400000);
        set_icon_state (Contents_win, 40, 0x00400000, 0x00400000);
      }
      break;

    case 37:			// Chapter numbers for titles option
      select_icon (winhandle, 37, Contents_ChapterTitlesNumTemp = !Contents_ChapterTitlesNumTemp);
      if (Contents_ChapterTitlesNumTemp)
        set_icon_state (Contents_win, 38, 0x00000000, 0x00400000);
      else
        set_icon_state (Contents_win, 38, 0x00400000, 0x00400000);
      break;

    case 40:			// Page numbers for titles option
      select_icon (winhandle, 40, Contents_ChapterTitlesPagTemp = !Contents_ChapterTitlesPagTemp);
      break;

    default:                   // click anywhere else - don't do anything
      break;

  }
}

//---------------------------------------------------------------------------------------------

// Key press event handler for contents style window.

int Contents_key_press (int winhandle, int icon, int key)
{
  switch (key)
  {
    case 13:			// Return, simulate mouse click over Do Contents
      Contents_click(winhandle,1,4,0,0);
      break;

    case 27:			// Esc, simulate mouse click over Cancel
      Contents_click(winhandle,2,4,0,0);
      break;

    default:
      break;
   }

   return key;
}

//=============================================================================================




////////////////////////////////////////////////
// Contents Menu Functions                    //
////////////////////////////////////////////////

// Called when the Contents menu entry is clicked.

int Contents_command(int n, int sub)
{
  // Check if the file has changed.

  Contents_close(Contents_win);

  if(Contents_current_file!=currentfile())
  {
   Contents_current_file=currentfile(); // Get file handle
   Contents_load_document();			// Load options saved with document
  }

  Contents_current_view=currentview();

  Contents_win=create_windowparent("Con_Contents",viewhandle(currentview()));

  // Add handlers for when the user clicks on the contents styles window or when
  // s(he) hits a key with it displayed.

  addwindowhandler(0,Contents_win,"Contents_click");
  addwindowhandler(1,Contents_win,"Contents_key_press");
  addwindowhandler(7,Contents_win,"Contents_close");
  addwindowhandler(2,Contents_win,"Contents_close");

  Contents_StyleWin();				// Display the Contents window

  setactive(Contents_win);
}

//=============================================================================================






////////////////////////////////////////////////
// Event handlers.			      //
////////////////////////////////////////////////

// Called when a file is closed.

void Contents_file_closing(int user,int file)
{
 if(file==Contents_current_file)
 {
  Contents_close(Contents_win);

 }
}


//---------------------------------------------------------------------------------------------

// Called when a view is made active.

void Contents_view_active (int user,int view,int gain)
{
 if(view==Contents_current_view)
 {
  if(!gain)
  {
   Contents_close(Contents_win);
  }
 }
}


//---------------------------------------------------------------------------------------------

// Called when a view is closed.

void Contents_view_closed(int user,int view)
{
 if(view==Contents_current_view)
 {
  Contents_close(Contents_win);
 }
}


//=============================================================================================

////////////////////////////////////////////////
// Main Functions                             //
////////////////////////////////////////////////

void main(void)
{
  string help = "{CONTENTS_10}";

  Contents_load_defaults ();	// Set the default options

  // Create the menus and windows, assign to handles.

  Contents_stylemenu=create_menu("{CONTENTS_05}");


  // Add contents menu to applets menu.

  script_menu_initialise();             // Initialise the menu
  script_menu_insert("Contents_command","","","","{CONTENTS_00}");



  // Add handlers for files and views being closed and views made active.

  addeventhandler (0x02, 0, "Contents_file_closing");
  addeventhandler (0x06, 0, "Contents_view_closed");
  addeventhandler (0x215, 0, "Contents_view_active");

  // Define a macro to launch us off.

  if (isdefmacro("contents") == -1)
  {
    translate(help);
    defmacro3(0,"contents", "{Contents_command(0,0)}", "contents", help,"Contents");
  }
}
